package org.fortiss.tooling.kernel.internal;

import static org.eclipse.emf.ecore.xmi.XMLResource.OPTION_PROCESS_DANGLING_HREF;
import static org.eclipse.emf.ecore.xmi.XMLResource.OPTION_PROCESS_DANGLING_HREF_DISCARD;

import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IConstraintVerifier;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.KernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.introspection.items.ConstraintVerificationServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.model.constraints.ConstrainedWithChecksum;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsFactory;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraint;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

/**
 * Implementation of {@link IConstraintVerificationService}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 159EDD10E24A64448987075B5F8B7B14
 */
public final class ConstraintVerificationService extends
		EObjectAwareServiceBase<IConstraintVerifier<IConstraint>> implements
		IConstraintVerificationService {

	/** {@inheritDoc} */
	@Override
	public void startService() {
		KernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** Registers the given verifier with the service. */
	@SuppressWarnings("unchecked")
	public <T extends IConstraint> void registerConstraintVerifier(IConstraintVerifier<T> verifier,
			Class<T> constraintElementClass) {
		addHandler(constraintElementClass, (IConstraintVerifier<IConstraint>)verifier);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThe service provides verification of constraints, the information whether a constraint is up to date or not;" +
				"\nit also provides potential fixes in case the verification of a constraint failed or resulted in an error." +
				"\n\nThis service is fed with \"constraint verifiers\"." +
				"\nConstraint verifiers are responsible of the verification of objects of type IConstraint." +
				"\nVerifiers do not return a status, but instead modify the status stored in the constraint:" +
				"\nthis allows to remember the results of potentially costly verifications." +
				"\n\nThe default statuses are SUCCESS, FAIL, ERROR, or OUTDATED:" +
				"\n- SUCCESS means the verification succeeded" +
				"\n- FAIL means the verification failed" +
				"\n- ERROR means the verification could not be run (typically because an error happened)" +
				"\n- OUTDATED means the constrained items have changed and the constraint sould be verified. " +
				"\n  Unless you decide on purpose not to verify a constraint (e.g., because deactivated), " +
				"\n  the verification should not end up with an outdated status." +
				"\nYou are encouraged to define your own statuses (which must however inherit from the default ones)," +
				"\nfor instance to provide more information about failure (thus possibly allowing automatic fixes)." +
				"\n\nNotes on the implementation of a verifier:" +
				"\n- Use ConstraintsUtils.setVerificationStatus to set the status instead of using directly the setter." +
				"\n- The service always runs verifications in a Command, so no need to do it inside the verifier. " +
				"\n\nThe service extension point is '" + getExtensionPointName() + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public void verify(IConstraint constraint) {
		IConstraintVerifier<IConstraint> verifier = getFirstVerifier(constraint);
		if(verifier != null) {
			ITopLevelElement modelContext =
					IPersistencyService.getInstance().getTopLevelElementFor(constraint);
			// It can happen that <code>modelContext</code> is null, e.g., if the verification is
			// triggered while the constraint is actually being removed.
			if(modelContext == null) {
				return;
			}
			modelContext.runAsCommand(new Runnable() {
				@Override
				public void run() {
					// We update the checksums before verification to avoid detecting some fake
					// changes during the verification.
					updateChecksums(constraint);
					verifier.verify(constraint);
					// And we also update the checksums after in case the verification had some side
					// effects...
					updateChecksums(constraint);
					constraint.getVerificationStatus().setConstraint(constraint);
				}
			});
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<IFix> fixes(IConstraintVerificationStatus status) {
		List<IFix> fixes = null;
		if(status != null) {
			IConstraintVerifier<IConstraint> verifier = getFirstVerifier(status.getConstraint());
			if(verifier != null) {
				fixes = verifier.fixes(status);
			}
		}
		return fixes != null ? fixes : new ArrayList<IFix>();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isUpToDate(IConstraint constraint) {
		for(ConstrainedWithChecksum cwc : constraint.getConstrainedsWithChecksum()) {
			BigInteger computeCheckSum = computeCheckSum(cwc.getConstrained(), constraint);
			if(computeCheckSum == null || !computeCheckSum.equals(cwc.getChecksum())) {
				return false;
			}
		}
		return true;
	}

	/**
	 * @param constraint
	 *            Updates all the checksums of <code>constraint</code>.
	 */
	private void updateChecksums(IConstraint constraint) {
		List<ConstrainedWithChecksum> newCwcs = new ArrayList<ConstrainedWithChecksum>();
		for(ConstrainedWithChecksum cwc : constraint.getConstrainedsWithChecksum()) {
			ConstrainedWithChecksum newCwc =
					ConstraintsFactory.eINSTANCE.createConstrainedWithChecksum();
			newCwc.setConstrained(cwc.getConstrained());
			newCwc.setChecksum(computeCheckSum(cwc.getConstrained(), constraint));
			newCwcs.add(newCwc);
		}
		EcoreUtils.disableNotificationsOf(constraint);
		constraint.getConstrainedsWithChecksum().clear();
		constraint.getConstrainedsWithChecksum().addAll(newCwcs);
		EcoreUtils.enableNotificationsOf(constraint);
	}

	/**
	 * Flag to print objects which are actually checksummed in various files.
	 * This is useful since it is essential to control very precisely how checksums are computed.
	 * Should however only be set to <code>true</code> on a local machine, not commited!
	 */
	private final boolean DEBUG = false;

	/** Number used to annotate the traces. */
	private int debug_trace_number = 0;

	/**
	 * @param constrained
	 * @param constraint
	 * @return the checksum of <code>constrained</code>. Note that <code>constraint</code> is
	 *         necessary because it has a potential impact on what is relevant for the checksum or
	 *         not.
	 */
	protected BigInteger computeCheckSum(IConstrained constrained, IConstraint constraint) {
		if(constrained == null) {
			return null;
		}
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		final ObjectOutputStream oos;
		final EObject hashRelevant = getChecksumRelevantEObject(constrained, constraint);
		try {
			final ResourceSet rset = new ResourceSetImpl();
			oos = new ObjectOutputStream(baos);
			Resource res = rset.createResource(URI.createURI("temp"));
			res.getContents().add(hashRelevant);
			Map<String, String> options = new HashMap<>();

			// The option below is to allow computing the hashsum of an incomplete ecore sub-model,
			// which is wanted in case parts of the model which are not checksum relevant are still
			// referred to.
			options.put(OPTION_PROCESS_DANGLING_HREF, OPTION_PROCESS_DANGLING_HREF_DISCARD);

			res.save(oos, options);
			if(DEBUG) {
				String fileName = "/home/aravantinos/temp/constraints/temp" + debug_trace_number++;
				String nameConstraint =
						constraint instanceof INamedElement ? ((INamedElement)constraint).getName()
								: constraint.toString();
				String nameConstrained =
						constrained instanceof INamedElement ? ((INamedElement)constrained)
								.getName() : constrained.toString();
				FileOutputStream file =
						new FileOutputStream(fileName + "_" + nameConstraint + "_" +
								nameConstrained + ".xml");
				ObjectOutputStream oos2 = new ObjectOutputStream(file);
				res.save(oos2, options);
				oos2.close();
			}
			oos.close();
			MessageDigest m = MessageDigest.getInstance("SHA1");
			m.update(baos.toByteArray());
			res.getContents().remove(hashRelevant);
			res.delete(options);
			return new BigInteger(1, m.digest());
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * @param constrained
	 * @param constraint
	 * @return The object to be used to compute the check sum.
	 *         Generally, all verification statuses and checksums are removed. Each verifier can
	 *         also provide per-constraint adaptations.
	 */
	protected EObject getChecksumRelevantEObject(IConstrained constrained, IConstraint constraint) {
		EObject res = EcoreUtils.copy(constrained);
		IConstraintVerifier<IConstraint> verifier = getFirstVerifier(constraint);
		if(verifier != null) {
			verifier.removeConstraintIrrelevantContent(res);
		}
		return res;
	}

	/**
	 * @param clazz
	 * @return the first registered verifier which can verify a constraint of class
	 *         <code>clazz</code>.
	 */
	protected IConstraintVerifier<IConstraint> getFirstVerifier(Class<?> clazz) {
		List<IConstraintVerifier<IConstraint>> list = getRegisteredHandlers(clazz);
		if(list == null || list.isEmpty()) {
			LoggingUtils.error(ToolingKernelActivator.getDefault(),
					String.format("Cannot find verifier for constraint %s", clazz));
			return null;
		}
		// get(0) because we know the list is not empty and because we want the first element
		return list.get(0);
	}

	/**
	 * @param constraint
	 * @return The first registered verifier which can verify <code>constraint</code>.
	 */
	protected IConstraintVerifier<IConstraint> getFirstVerifier(IConstraint constraint) {
		return getFirstVerifier(constraint.getClass());
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return "org.fortiss.tooling.kernel.constraintVerifier";
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return "constraintVerifier";
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return "constraintVerifier";
	}

	/** {@inheritDoc} */
	@Override
	protected String getClassAttribute() {
		return "constraint";
	}

	/** {@inheritDoc} */
	@Override
	protected String getClassConfigurationElement() {
		return "constraint";
	}

	/** {@inheritDoc} */
	@Override
	public void setConstrained(IConstraint constraint, IConstrained constrained, int index) {
		EList<ConstrainedWithChecksum> cwcs = constraint.getConstrainedsWithChecksum();
		if(cwcs.size() <= index) {
			for(int i = cwcs.size(); i <= index; i++) {
				cwcs.add(i, ConstraintsFactory.eINSTANCE.createConstrainedWithChecksum());
			}
		}
		if(!constrained.getConstraints().contains(constraint)) {
			constrained.getConstraints().add(constraint);
		}
		ConstrainedWithChecksum cwc = cwcs.get(index);
		cwc.setConstrained(constrained);
		cwc.setChecksum(computeCheckSum(constrained, constraint));
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Constraint Verification Service";
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new ConstraintVerificationServiceIntrospectionDetailsItem(handlerMap);
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<?>> getConstraintTypes() {
		return this.handlerMap.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public <T extends IConstraint> String getDescription(Class<T> constraintEltClass) {
		IConstraintVerifier<IConstraint> verifier = getFirstVerifier(constraintEltClass);
		return verifier != null ? verifier.getDescription() : null;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Set<Class<?>> getDevelopmentProcessConstraintTypes() {
		return this.handlerMap.keySet();
	}
}
